
/****************************************************************************
** File: SAP.CPP
**
**	Description:
**
**		The following code is an application skeleton that provides
**		sap services for OS/2 32 bit compilers.  It spawns a thread
**		to handle sapping every 60 seconds, as well as a separate thread
**		to respond to general service queries and nearest service queries.
**		It is written to support a single sapping service, although all
**		of the functions should be reentrant.  So, with a little work,
**		this should easily be able to handle multiple sapping devices on
**		one workstation.  The two functions to be called in this example
**		are AdvertiseService and ShutdownSAP.  All other fuctions are
**		subsidiary to these functions.
**
**	Disclaimer:
**
**	     Novell, Inc. makes no representations or warranties with respect to
**	     any NetWare software, and specifically disclaims any express or
**	     implied warranties of merchantability, title, or fitness for a
**	     particular purpose.
**
**	     Distribution of any NetWare software is forbidden without the
**	     express written consent of Novell, Inc.  Further, Novell reserves
**	     the right to discontinue distribution of any NetWare software.
**
**	     Novell is not responsible for lost profits or revenue, loss of use
**	     of the software, loss of data, costs of re-creating lost data, the
**	     cost of any substitute equipment or program, or claims by any party
**	     other than you.  Novell strongly recommends a backup be made before
**	     any software is installed.   Technical support for this software
**	     may be provided at the discretion of Novell.
**
**	Programmers:
**
**	     Ini	Who					Firm
**	     -------------------------------------------------------------------
**	     DRS	Dan Stratton			Novell Developer Support
**
**	History:
**
**	     When		Who	What
**	     -------------------------------------------------------------------
**	     07-01-94		DRS	First code.
**           02-01-95		DRS	Added General Service Query responses,
**					as well as making it compatible with CSET/++
**	     02-24-95		DRS	Fixed some NSQ response problems.  Works with
**					DOS QueryServices call now.
**
*/



#define NWOS2
#define IS32BIT
/* for now, just allow to run with either Borland or IBM */
#if defined	(__BORLANDC__)
	#define BCPP
#elif (defined (__IBMC__) || defined(__IBMCPP__))
	#define CSET2
#endif

#define INCL_DOSPROCESS
#define INCL_DOSSEMAPHORES

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <os2.h>
#include <conio.h>
#include <stddef.h>
#include <process.h>

#ifdef __cplusplus
	extern "C"{
#endif
#include <ipxcalls.h>
#include <nwmisc.h>
#ifdef __cplusplus
	}
#endif


#define SAP_SOCKET           	0x0452
#define SAP_PACKET_TYPE         4
#define SAPSLEEPTIME		60000 /* 60 seconds sleep time */
#define GSQBUFFSIZE		512
#define MAXSAPTHREADS	10    /* arbitrary for now, it only handles one sapping server */




/* ****************************************************************
			Typedefs				    */

typedef struct IPXAddress
  {
    BYTE    network[4];                /* high-low */
    BYTE    node[6];                   /* high-low */
    BYTE    socket[2];                 /* high-low */
  }IPXAddress;

typedef struct SAPHeader
  {
     WORD          checksum;             /* high-low 1's complement */
     WORD          length;               /* high-low unsigned int */
     BYTE          transportControl;
     BYTE          packetType;
     IPXAddress    destination;
     IPXAddress    source;
     WORD          SAPPacketType;        /* 2 or 4 */
     WORD          serverType;           /* assigned by Novell */
     BYTE          serverName[48];       /* Server name */
     IPXAddress    serverAddress;        /* server internetwork address */
     WORD          interveningNetworks;  /* # of networks packet must traverse */
  } SAPHeader;

typedef struct
  {
    IPX_HEADER    ipxHdr;
    WORD          ResponseType;         /* HI - LO  */
    WORD          ServerType;           /* HI - LO  */
    BYTE          ServerName[48];
    BYTE          Network[4];           /* hi - lo  */
    BYTE          Node[6];              /* hi - lo  */
    BYTE          Socket[2];            /* hi - lo  */
    WORD          InterveningNetworks;  /* hi - lo  */
  } SAP;


typedef struct
  {
    WORD	  queryType;
    WORD	  serverType;
  } QUERY_PACKET;

typedef struct
  {
     USHORT 	  ipxSocket;
     WORD 	  serverType;
     BYTE	  serverName[48];
  } ADV_SERVICE;



/* Function prototypes  */

#ifdef CSET2
	#pragma stack16(2048)   /* limit 16bit stack allocated on thread stacks */
	void  _Export _Optlink NW_AdvertiseSAP(void *);
	void  _Export _Optlink GSQthreadCode(void *);
#else
	void   NW_AdvertiseSAP(void *advServ);
#endif

 int  	SetupSAPPacket(IPX_ECB *ipxEcb, SAP *sap, BYTE *dest, ADV_SERVICE *advServ, WORD responseType);
 void 	GSQthreadCode(void *advServ);
 void 	GSQSetup(IPX_ECB *gsqECB, IPX_HEADER *gsqHeader, char *gsqBuff);
 void	main(void);
 int    AdvertiseService(WORD serverType, char *serverName, USHORT ipxSocket);
 int    ShutdownSAP(USHORT ipxSocket);


/* Global Variables */

 int	currSapThreads = 0;
 USHORT sapSocket;
 BOOL   killAllThreads = FALSE;
 TID  	sapThread[MAXSAPTHREADS];
 HEV	gsqCritSec;
 HEV	sapCritSec;


 /* ************************************************************************
 GSQthreadCode

 This code is started from AdvertiseService.  It begins a receive loop that
 that gets all SAP traffic, checks to see if the packet is a General Service
 Query or Nearest Service Query, and that it is asking for the specified
 server type.  If so, the code responds directly to the requesting  station.

 *************************************************************************** */

void GSQthreadCode(void *advServ)
{
	ULONG 		timeOut = 1000L;
	int 		ccode = 0;
	BOOL 		killThisThread = FALSE;
	WORD 		gsqPacket;
	WORD		nsqPacket;
	WORD 		queryServer;
	IPX_ECB     	ipxEcb;
	BYTE        	dest[10];
	SAP            	sap;
	IPX_HEADER    	gsqHeader;
	IPX_ECB	gsqECB;
	char 		gsqBuff[GSQBUFFSIZE];
	ULONG 		semTimeOut = 1000L;
	ULONG 		rc;
	ULONG		ulPostCtl;
	WORD		responseType;

	memset((void *)&sap, 0, sizeof(SAP));


	gsqPacket = NWWordSwap(0x01);	 	 /* General Service Query */
	nsqPacket = NWWordSwap(0x03);	 	 /* Nearest Service Query */
	queryServer = NWWordSwap(((ADV_SERVICE *)advServ)->serverType);
	memset((char *)&gsqECB,0,sizeof(IPX_ECB));
	memset((char *)&gsqHeader,0,sizeof(IPX_HEADER));
	memset((char *)&gsqBuff,0,sizeof(gsqBuff));
	gsqECB.fragList[0].fragAddress = &gsqHeader;
	gsqECB.fragList[0].fragSize = sizeof(IPX_HEADER);
	gsqECB.fragList[1].fragAddress = &gsqBuff;
	gsqECB.fragList[1].fragSize = GSQBUFFSIZE;
	gsqECB.fragCount = 2;
	gsqECB.next = gsqECB.prev = NULL;


	do{
	  rc=DosWaitEventSem(gsqCritSec, semTimeOut);
	  if (rc!=0)
	  {
	    printf("DosWaitEventSem returned error %ld in GSQ\n", rc);
	    DosExit(EXIT_THREAD, 1);
	  }


	  rc = DosResetEventSem(gsqCritSec, &ulPostCtl);
	  if (rc!=0)
	  {
	    printf("DosResetEventSem returned error %ld in GSQ\n", rc);
	    DosExit(EXIT_THREAD, 1);
	  }

	  ccode = IpxReceive(sapSocket,timeOut,&gsqECB);

	  rc = DosPostEventSem(gsqCritSec);
	  if (rc!=0)
	  {
	    printf("DosPostEventSem returned error %ld in GSQ\n", rc);
	    DosExit(EXIT_THREAD, 1);
	  }

	  switch(ccode)     /* check for success or error condition */
	  {
		case 0x0000:
		     /* Is this a GSQ/NSQ for our type of server */
		     if (  ((((QUERY_PACKET *)&gsqBuff)->queryType == gsqPacket) ||
		       (((QUERY_PACKET *)&gsqBuff)->queryType == nsqPacket)) &&
		       (NWWordSwap(((QUERY_PACKET *)&gsqBuff)->serverType) == queryServer))
			{
			  memcpy((void *)(&sap.ipxHdr.destNet), (void *)(&gsqHeader.sourceNet), 12);
			  memcpy(dest, (void *)(&sap.ipxHdr.destNet), 10);

			  /*
			     This is just so we can see when General Service Queries and Nearest
			     Service Queries are received.
			  */
			  if( ((QUERY_PACKET *)&gsqBuff)->queryType == nsqPacket)
			  {
			   responseType = 0x0400;
			   printf("NSQ received from %08X / %02X%02X%02X%02X%02X%02X / %04X\n",
				NWLongSwap(sap.ipxHdr.destNet),sap.ipxHdr.destNode[0],
				sap.ipxHdr.destNode[1], sap.ipxHdr.destNode[2],
				sap.ipxHdr.destNode[3], sap.ipxHdr.destNode[4],
				sap.ipxHdr.destNode[5],
				NWWordSwap(sap.ipxHdr.destSocket));
			  }
			  else
			  {
			    responseType = 0x0200;
			    printf("GSQ received from %08X / %02X%02X%02X%02X%02X%02X / %04X\n",
				NWLongSwap(sap.ipxHdr.destNet),sap.ipxHdr.destNode[0],
				sap.ipxHdr.destNode[1], sap.ipxHdr.destNode[2],
				sap.ipxHdr.destNode[3], sap.ipxHdr.destNode[4],
				sap.ipxHdr.destNode[5],
				NWWordSwap(sap.ipxHdr.destSocket));
			   }
			  fflush(stdout);
			  ccode = SetupSAPPacket(&ipxEcb, &sap, dest, (ADV_SERVICE *)advServ,responseType);
			  if(ccode)
			  {
				printf("\nError setting up SAP packet, terminating thread");
				killThisThread = TRUE;
			  }

			  DosWaitEventSem(gsqCritSec, semTimeOut);
			  rc = DosResetEventSem(gsqCritSec, &ulPostCtl);
			  if (rc!=0)
			  {
			    printf("DosResetEventSem returned error %ldin GSQ #2\n", rc);
			    DosExit(EXIT_THREAD, 1);
			  }

			  ccode = IpxSend(gsqHeader.sourceSocket, (IPX_ECB NWPTR)&ipxEcb);
			  rc = DosPostEventSem(gsqCritSec);
			  if (rc!=0)
			  {
			    printf("DosPostEventSem returned error %ld in GSQ #2\n", rc);
			    DosExit(EXIT_THREAD, 1);
			  }

			  if (ccode!=0)
			  {
				printf("IpxSend returned error %ld\n", ccode);
				killThisThread = TRUE;
			  }
			}
			break;
		case 0x9001:  break;
		case 0x8007:  killThisThread = TRUE;
		       break;
		case 0x8002:  printf("\nError: Bad Packet Received on thread %ld",_threadid);
		       break;
		case 0x8006:  printf("\nError: Receive Overflow, buffer too small on thread %ld",_threadid);
		       break;
		case 0x9004:  printf("\nError: Socket not open, Terminating thread %ld",_threadid);
		       killThisThread = TRUE;
		       break;
		default:      printf("\nUnknown error during IPXReceive.  Terminating thread %ld, ccode %04X",_threadid,ccode);
		       killThisThread = TRUE;
		       break;
	  }
      /* Do one IPXRECEIVE, then go to sleep until the next timeslice */
      DosSleep(0);
   }while (!killThisThread && !killAllThreads);

   _endthread();
}



/* ***********************************************************************


       SetupSapPacket

       This code does general setup for the Sapping and GSQ/NSQ threads.


   *********************************************************************** */

int SetupSAPPacket(IPX_ECB *ipxEcb, SAP *sap, BYTE *dest, ADV_SERVICE *advServ, WORD responseType)
{
  ULONG       	tm;             /* Round trip time */
  int 		ccode = 0;

  tm = 0;


  ipxEcb->next  = (IPX_ECB NWPTR)NULL;
  ipxEcb->prev  = (IPX_ECB NWPTR)NULL;
  ipxEcb->fragCount               = 1;
  ipxEcb->fragList[0].fragAddress = sap;
  ipxEcb->fragList[0].fragSize    = sizeof(SAP);

  sap->ipxHdr.packetType  = SAP_PACKET_TYPE;

  sap->ResponseType       = responseType;
  sap->ServerType         = NWWordSwap(advServ->serverType); 	/* Server type - LO-HI order */
  strcpy((char*)(sap->ServerName), (char *)advServ->serverName);/* Server Name */

      /* This is an old call, but we haven't added the new one to the thunked headers yet */
  ccode = IpxGetInternetworkAddress(sap->Network);
  if (ccode!=0)
  {
    printf("IpxGetInternetworkAddress returned error %ld", ccode);
    return (1);
  }


  sap->Socket[0]          =(BYTE) (NWWordSwap(advServ->ipxSocket) >>8);      /* HI byte of Service Socket */
  sap->Socket[1]          =(BYTE) (NWWordSwap(advServ->ipxSocket) & 0x00FF); /* LO byte of Service Socket  */
  sap->InterveningNetworks= NWWordSwap(0x0001);      /* LO-HI order - intially set to 1 */

  /* Get local target sets immediate address, etc.  */
  ccode = IpxGetLocalTarget(dest, ipxEcb, &tm);
  if (ccode!=0)
  {
    printf("IpxGetLocalTarget returned error %ld", ccode);
    return (1);
  }

  return (0);
}



/* ************************************************************************
 NW_AdvertiseSap

 This code is started from AdvertiseService.  It begins a send loop that
 advertises the SAP service every 60 seconds.

  *************************************************************************** */


void NW_AdvertiseSAP(void *advServ)
{


  IPX_ECB     	ipxEcb;
  int      	ccode;
  BYTE        	dest[10];
  SAP         	sap;
  ULONG		ulPostCtl;
  ULONG 	semTimeOut = 1000L;
  int		rc;
  WORD		responseType;

			      /* This is the destination of the SAP packet */
  memset(&dest, 0, 4);        /* SAP's are braoadcast to the local net only. */
  memset(&dest[4], 0xFF, 6);  /* Routers will then propagate this information */
  sap.ipxHdr.destNet    = 0x00000000;
  sap.ipxHdr.destNode[0]= 0xFF;
  sap.ipxHdr.destNode[1]= 0xFF;
  sap.ipxHdr.destNode[2]= 0xFF;
  sap.ipxHdr.destNode[3]= 0xFF;
  sap.ipxHdr.destNode[4]= 0xFF;
  sap.ipxHdr.destNode[5]= 0xFF;
  sap.ipxHdr.destSocket = NWWordSwap(SAP_SOCKET); /* turn socket into lo-hi */

  responseType = 0x0200;

  ccode = SetupSAPPacket(&ipxEcb, &sap, dest, (ADV_SERVICE *)advServ,responseType);

  if(ccode)
  {
    printf("\nError setting up SAP packet, terminating thread");
    DosExit(EXIT_THREAD, 1);
  }

  while(!killAllThreads)
  {

   DosWaitEventSem(sapCritSec, semTimeOut);
   rc = DosResetEventSem(sapCritSec, &ulPostCtl);
   if (rc!=0)
    {
     printf("DosResetEventSem returned error %ld in SAP\n", rc);
     DosExit(EXIT_THREAD, 1);
    }
   ccode = IpxSend( (((ADV_SERVICE *)advServ)->ipxSocket), (IPX_ECB NWPTR)&ipxEcb);

   rc=DosPostEventSem(sapCritSec);
   if (rc!=0)
    {
     printf("DosPostEventSem returned error %ld in SAP\n", rc);
     DosExit(EXIT_THREAD, 1);
    }
   if (ccode!=0)    {
     printf("IpxSend returned error %ld\n", ccode);
     DosExit(EXIT_THREAD, 1);
    }
    DosSleep(SAPSLEEPTIME);     /* Go to sleep until we need to SAP again */
  }
 }



/*  *******************************************************************

	AdvertiseService

	Call AdvertiseService to begin sapping.  Pass in the server type
	(this should be given out by Novell if it is a new one), server
	name, and the socket you opened that the server will communicate
	over.  This can be a dynamic socket or one that you have had assigned.

     ****************************************************************** */



int AdvertiseService(WORD serverType, char *serverName, USHORT ipxSocket)
{

  int 		ccode;
  ADV_SERVICE   *advServ;
  ULONG		rc;
  char          semName[20];

  advServ = (ADV_SERVICE *)calloc(1,sizeof(ADV_SERVICE));
  if (advServ == NULL)
  {
    printf("calloc failed\n");
    return 1;
  }

  sapSocket = NWWordSwap(SAP_SOCKET); /* Open SAP socket */
  ccode = IpxOpenSocket(&sapSocket);
  if (ccode!=0)
  {
    printf("IpxOpenSocket returned error %04Xon socket %04X", ccode,NWWordSwap(sapSocket));
    return 1;
  }
  strcpy(semName,"\\sem32\\gsqCrit");
  rc=DosCreateEventSem((PSZ)semName, &sapCritSec, 0, TRUE);
  if (rc!=0)
  {
    printf("DosCreateEventSem returned error %d\n", rc);
    return 1;
  }

  strcpy(semName,"\\sem32\\sapCrit");
  rc=DosCreateEventSem((PSZ)semName, &gsqCritSec, 0, TRUE);
  if (rc!=0)
  {
    printf("DosCreateEventSem returned error %d\n", rc);
    return 1;
  }

  advServ->ipxSocket =  ipxSocket;                   /* Put appropriate data elements in */
  advServ->serverType = serverType;                  /* a structure that we can pass through */
  strcpy((char *)(advServ->serverName) , serverName);  /* beginThread.  It accepts a single void *. */

#ifdef CSET2
     if ((sapThread[currSapThreads++] = _beginthread(NW_AdvertiseSAP, NULL, 8192,(void *)advServ)) < 0)
#else
     if ((sapThread[currSapThreads++] = _beginthread(NW_AdvertiseSAP, 8192, (void *)advServ)) < 0)
#endif
     {
	  printf("Unable to create AdvertiseSapthread,errno = %d\n",errno);
	  return 1;
      }


 /*  Create a thread to respond to general service queries */


#ifdef CSET2
      if ((sapThread[currSapThreads++] = _beginthread(GSQthreadCode, NULL, 8192,(void *)advServ) ) < 0)
#else
      if ((sapThread[currSapThreads++] = _beginthread(GSQthreadCode, 8192, (void *)advServ) ) < 0)
#endif
	{
	   printf("Unable to create GSQ thread,errno = %d\n",errno);
	   return 1;
	}
     return 0;

}


/* *************************************************************************
	ShutdownSAP

	Use this function to shut down the SAP and GSQ/NSQ threads.  The
	semaphores ensure that we are not in the middle of a NetWare call
	when we terminate (that would be bad).  It also closes the open
	sockets.  This function will need some logic changes if you want to
	provide support for multiple sapping devices.

  ************************************************************************** */


int ShutdownSAP(USHORT ipxSocket)
{

  int 		i=0;
  int		ccode = 0;
  ULONG		ulPostCtl;
  ULONG		semTimeOut=1000; /* arbitrary time-out value */

  killAllThreads = TRUE;
  IpxCloseSocket(ipxSocket);
  IpxCloseSocket(sapSocket);

  DosWaitEventSem(sapCritSec, semTimeOut);
  DosResetEventSem(sapCritSec,&ulPostCtl);
  DosWaitEventSem(gsqCritSec, semTimeOut);
  DosResetEventSem(gsqCritSec,&ulPostCtl);

  for(i=0;i<=currSapThreads; ++i)
  ccode = DosKillThread(sapThread[i]);
   if (ccode!=0 && ccode!=309)
   {
     printf("DosKillThread returned error %d\n", ccode);
     return 1;
   }

return 0;
}


/* ************************************************************************
				main
   ************************************************************************ */

void main ( void ) {


  USHORT      	ipxSocket;
  int		ccode = 0;
  WORD 		serverType;
  char		serverName[48];



/*
    The following few lines are where you set up your server name, socket,
    and server type.  Set this for your own types and names.
*/
  serverType = 0x0ABC;
  strcpy(serverName, "TEST_Server");
  ipxSocket = 0x0000;
  ccode = IpxOpenSocket(&ipxSocket);
  if (ccode!=0)
  {
    printf("IpxOpenSocket returned error %04X on socket %04X", ccode, NWWordSwap(ipxSocket));
    exit(1);
  }


/*  Call AdvertiseService to begin sapping and responding to GSQ/NSQ queries */

  ccode = AdvertiseService(serverType, serverName, ipxSocket);
  if (ccode != 0)
  {
    printf("AdvertiseService returned error %04X\n", ccode);
    exit(1);
  }

  /*  The main program would go after this point. */

  while(!kbhit())  /* Spin loop for this example */
  {
    DosSleep(500l);  /* give up 500 ms - arbitrary value just for this example */
  }


  ccode = ShutdownSAP(ipxSocket);
  if (ccode != 0)
  {
    printf("ShutdownSap returned error %04X\n", ccode);
    exit(1);
  }

} /* end main */
